#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <sys/time.h>
#include <gsl/gsl_rng.h>
#include "island_ga.h"
#include "omp.h"

#define NUMBER_ISLANDS      8
#define POPULATION_SIZE     (1600/NUMBER_ISLANDS)
#define NUMBER_CHROMOSOMES  150
#define ALLELES             4
#define MUTATION_RATE       0.0035
#define MAXIMUM_GENERATIONS 10000
#define RUNS                15

static int population_size = POPULATION_SIZE; 

Organism*** initialize() {
  const gsl_rng_type * T;
  gsl_rng * r;
  int i, n = 10;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  Organism*** population_buffers = malloc(NUMBER_ISLANDS*2*sizeof(Organism**));
  for(int i=0;i<NUMBER_ISLANDS*2;i+=2) {
    // First population buffer filled with random organisms
    population_buffers[i] = malloc(POPULATION_SIZE*sizeof(Organism*));
    for(int j=0;j<POPULATION_SIZE; j++) {
      population_buffers[i][j] = create_organism(r);
    }
    // Secod population buffer filled with empty organisms
    population_buffers[i+1] = malloc(POPULATION_SIZE*sizeof(Organism*));
    for(int j=0;j<POPULATION_SIZE;j++) {
      population_buffers[i+1][j] = malloc(sizeof(Organism));
      population_buffers[i+1][j]->geneotype = malloc(NUMBER_CHROMOSOMES*sizeof(int));
    }
  }
  return population_buffers;
}

Organism* create_organism(gsl_rng* r) {
  Organism* organism = malloc(sizeof(Organism));
  organism->geneotype = malloc(NUMBER_CHROMOSOMES*sizeof(int));
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    organism->geneotype[i] = generate_random_chromosome(r);
  }
  organism->fitness = 0;
  return organism;
}

int generate_random_chromosome(gsl_rng* r) {
  return gsl_rng_uniform_int(r, ALLELES);
}

double evaluate_organisms(Organism** population) {
  double best_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    evaluate_organism(population[i]);
    if(best_fitness<population[i]->fitness) best_fitness = population[i]->fitness;
  }
  return best_fitness;
}

void evaluate_organism(Organism* organism) {
  int tally = 1;
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    if(organism->geneotype[i]!=1) tally++;
  }
  organism->fitness = 1./(double)tally;
}

Organism** produce_next_generation(Organism** current_population, Organism** next_population, gsl_rng* r) {
  apply_crossover(current_population, next_population, r);
  apply_mutations(next_population, r);
}

void apply_crossover(Organism** current_population, Organism** next_population, gsl_rng* r) {
  double total_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    total_fitness+=current_population[i]->fitness;
  }
  for(int i=0;i<POPULATION_SIZE;i+=2) {
    int parrent1 = select_one_organism(current_population, total_fitness, r);
    int parrent2 = select_one_organism(current_population, total_fitness, r);
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      int selector = gsl_rng_uniform_int(r, 2);
      next_population[i]->geneotype[j] = current_population[selector==0?parrent1:parrent2]->geneotype[j];
      next_population[i+1]->geneotype[j] = current_population[selector==0?parrent2:parrent1]->geneotype[j];
    }
  }
}

int select_one_organism(Organism** population, double total_fitness,  gsl_rng* r) {
  double running_total = 0;
  double selected_point = gsl_rng_uniform(r)*total_fitness;
  for(int i=0;i<POPULATION_SIZE;i++) {
    running_total+=population[i]->fitness;
    if(running_total>selected_point) return i;
  }
  return 0;
}

void apply_mutations(Organism** population, gsl_rng* r) {
  for(int i=0;i<POPULATION_SIZE;i++) {
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      if(gsl_rng_uniform(r) < MUTATION_RATE) {
        population[i]->geneotype[j] = generate_random_chromosome(r);
      }
    }
  }
}

void print_orgamism(Organism* organism) {
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    printf("%i ", organism->geneotype[i]);
  }
  printf("[%f]\n", organism->fitness);
}


#pragma omp task inout([population_size]current_population, [population_size]next_population) copy_deps
void run_one_generation(Organism** current_population, Organism** next_population, int generation, double* best_fitness, int island_id, gsl_rng* r) {
    best_fitness[island_id] = evaluate_organisms(current_population);
    if(best_fitness[island_id] < 0) {
      for(int i=0;i<POPULATION_SIZE;i++) {
        print_orgamism(current_population[i]);
      }
    }
    produce_next_generation(current_population, next_population, r);
    evaluate_organisms(next_population);
}

gsl_rng* create_rng(int seed, bool seed_time) {
  const gsl_rng_type * T;
  gsl_rng * r;
  int i, n = 10;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  struct timeval time;
  gettimeofday( &time, NULL );
  if(seed_time) {
    gsl_rng_set(r, seed*(int)time.tv_usec);
  } else {
    gsl_rng_set(r, seed);
  }
  return r;
}

int run() {
  gsl_rng** island_rngs = malloc(NUMBER_ISLANDS*sizeof(gsl_rng*));
  for(int i=0;i<NUMBER_ISLANDS;i++) {
    island_rngs[i] = create_rng(i+1, false);
  }
  Organism*** populations = initialize();
  int final_generations[NUMBER_ISLANDS];
  double best_fitness = 0;
  int generation = 0;
  int island_ids[16] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
  while(best_fitness<0.999999 && generation < MAXIMUM_GENERATIONS) {
    for(int j=0;j<10;j++) {
      double* local_best_fitnesses = malloc(NUMBER_ISLANDS*sizeof(double));
      for(int i=0;i<NUMBER_ISLANDS;i++) {
        run_one_generation(populations[2*i+(generation%2)], populations[2*i+((generation+1)%2)], generation, local_best_fitnesses, island_ids[i], island_rngs[i]);
      }
#pragma omp taskwait
      for(int i=0;i<NUMBER_ISLANDS;i++) if(local_best_fitnesses[i]>best_fitness) best_fitness=local_best_fitnesses[i];
      free(local_best_fitnesses);
      generation++;
    }

    for(int j=0;j<NUMBER_ISLANDS;j++) {
      Organism** population1 = populations[2*j+((generation+1)%2)];
      Organism** population2 = populations[2*((j+1)%NUMBER_ISLANDS)+((generation+1)%2)];
      double total_fitness1 = 0;
      double total_fitness2 = 0;
      for(int i=0;i<POPULATION_SIZE;i++) {
        total_fitness1+=population1[i]->fitness;
        total_fitness2+=population2[i]->fitness;
      }
      int from_population1 = select_one_organism(population1, total_fitness1, island_rngs[j]);
      int from_population2 = select_one_organism(population2, total_fitness2, island_rngs[(j+1)%NUMBER_ISLANDS]);
      if(population1[from_population1]->fitness>population2[from_population2]->fitness) {
        population2[from_population2] = population1[from_population1];
      } else {
        population1[from_population1] = population2[from_population2];
      }
    }
  }
  printf("Final generation: %d\n", generation);
  return generation;
}

void print_timestamp() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long long millisecondsSinceEpoch =
    (unsigned long long)(tv.tv_sec) * 1000 +
    (unsigned long long)(tv.tv_usec) / 1000;
  printf("%llu\n", millisecondsSinceEpoch);
}

int main() {
  print_timestamp();
  for(int i=0;i<RUNS;i++) {
    run();
    print_timestamp();
  }

  return 0;
}
