#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <gsl/gsl_rng.h>
#include "ga.h"
#include "omp.h"
#include <CL/cl.h>
#define MEM_SIZE (128)
#define MAX_SOURCE_SIZE (0x100000)

#define NUMBER_TASKS          1
#define POPULATION_SIZE       1600
#define TASK_POPULATION_SIZE  POPULATION_SIZE/NUMBER_TASKS
#define NUMBER_CHROMOSOMES    150
#define ALLELES               4
#define MUTATION_RATE         0.0035
#define MAXIMUM_GENERATIONS   10000
#define RUNS                  1

int* initialize_population() {
  const gsl_rng_type * T;
  gsl_rng * r;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  int* population = malloc(POPULATION_SIZE*NUMBER_CHROMOSOMES*sizeof(int));
  for(int i=0;i<POPULATION_SIZE;i++) {
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      population[i*NUMBER_CHROMOSOMES+j] = generate_random_chromosome(r);
    }
  }
  return population;
}

Organism* create_organism(gsl_rng* r) {
  Organism* organism = malloc(sizeof(Organism));
  organism->geneotype = malloc(NUMBER_CHROMOSOMES*sizeof(int));
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    organism->geneotype[i] = generate_random_chromosome(r);
  }
  organism->fitness = 0;
  return organism;
}

int generate_random_chromosome(gsl_rng* r) {
  return gsl_rng_uniform_int(r, ALLELES);
}

double evaluate_organisms(int* population, double* fitness_values) {
  double best_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    fitness_values[i] = evaluate_organism(population, i);
    if(best_fitness<fitness_values[i]) best_fitness = fitness_values[i];
  }
  return best_fitness;
}

double evaluate_organisms_gpu(int* population, double* fitness_values) {
  cl_device_id device_id = NULL;
  cl_context context = NULL;
  cl_command_queue command_queue = NULL;
  cl_mem memobj = NULL;
  cl_program program = NULL;
  cl_kernel kernel = NULL;
  cl_platform_id platform_id = NULL;
  cl_uint ret_num_devices;
  cl_uint ret_num_platforms;
  cl_int ret;
  cl_int _err;

  char string[MEM_SIZE];

  FILE *fp;
  char fileName[] = "./ga.cl";
  char *source_str;
  size_t source_size;

  /* Load the source code containing the kernel*/
  fp = fopen(fileName, "r");
  if (!fp) {
    fprintf(stderr, "Failed to load kernel.\n");
    exit(1);
  }
  source_str = (char*)malloc(MAX_SOURCE_SIZE);
  source_size = fread(source_str, 1, MAX_SOURCE_SIZE, fp);
  fclose(fp);

  /* Get Platform and Device Info */
  ret = clGetPlatformIDs(1, &platform_id, &ret_num_platforms);
  ret = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_DEFAULT, 1, &device_id, &ret_num_devices);

  /* Create OpenCL context */
  context = clCreateContext(NULL, 1, &device_id, NULL, NULL, &ret);

  /* Create Command Queue */
  command_queue = clCreateCommandQueue(context, device_id, 0, &ret);

  /* Create Kernel Program from the source */
  program = clCreateProgramWithSource(context, 1, (const char **)&source_str,
      (const size_t *)&source_size, &ret);

  /* Build Kernel Program */
  ret = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);

  /* Create OpenCL Kernel */
  kernel = clCreateKernel(program, "ga", &ret);

  cl_mem input_buffer;
  input_buffer = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(int)*POPULATION_SIZE*NUMBER_CHROMOSOMES, NULL, &_err);

  cl_mem output_buffer;
  output_buffer = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(double)*POPULATION_SIZE, NULL, &_err);

  int number_of_chromosomes = NUMBER_CHROMOSOMES;

  clSetKernelArg(kernel, 0, sizeof(input_buffer), &input_buffer);
  clSetKernelArg(kernel, 1, sizeof(output_buffer), &output_buffer);
  clSetKernelArg(kernel, 2, sizeof(number_of_chromosomes), &number_of_chromosomes);

  for (int i=0; i<POPULATION_SIZE*NUMBER_CHROMOSOMES; i++) {
    clEnqueueWriteBuffer(command_queue, input_buffer, CL_FALSE, i*sizeof(int), sizeof(int), population+i, 0, NULL, NULL);
  }
  cl_event kernel_completion;
  size_t global_work_size[1] = { POPULATION_SIZE };
  clEnqueueNDRangeKernel(command_queue, kernel, 1, NULL, global_work_size, NULL, 0, NULL, &kernel_completion);
  clWaitForEvents(1, &kernel_completion);
  clReleaseEvent(kernel_completion);

  for (int i=0; i<POPULATION_SIZE; i++) {
    double data;
    clEnqueueReadBuffer(command_queue, output_buffer, CL_TRUE, i*sizeof(double), sizeof(double), fitness_values+i, 0, NULL, NULL);
  }
  double best_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    if(best_fitness<fitness_values[i]) best_fitness=fitness_values[i];
  }

  /* Finalization */
  ret = clFlush(command_queue);
  ret = clFinish(command_queue);
  ret = clReleaseKernel(kernel);
  ret = clReleaseProgram(program);
  ret = clReleaseMemObject(memobj);
  ret = clReleaseCommandQueue(command_queue);
  ret = clReleaseContext(context);

  free(source_str);

  return best_fitness;
}

double evaluate_organism(int* population, int individ) {
  double fitness;
  for(int j=0;j<1;j++) {
    int tally = 1;
    for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
      if(population[individ*NUMBER_CHROMOSOMES+i]!=1) tally++;
    }
    fitness = 1./(double)tally;
  }
  return fitness;
}

void produce_next_generation(int* current_population, int* next_population, double* fitness_values, int task_id, gsl_rng* r) {
  apply_crossover(current_population, next_population, fitness_values, task_id, r);
}

void apply_crossover(int* current_population, int* next_population, double* fitness_values, int task_id, gsl_rng* r) {
  double total_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    total_fitness+=fitness_values[i];
  }
  for(int i=0;i<POPULATION_SIZE/NUMBER_TASKS;i+=2) {
    int parrent1 = select_one_organism(fitness_values, total_fitness, r);
    int parrent2 = select_one_organism(fitness_values, total_fitness, r);
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      int selector = gsl_rng_uniform_int(r, 2);
      next_population[task_id*TASK_POPULATION_SIZE*NUMBER_CHROMOSOMES+i*NUMBER_CHROMOSOMES+j] = current_population[(selector==0?parrent1:parrent2)*NUMBER_CHROMOSOMES+j];
      next_population[task_id*TASK_POPULATION_SIZE*NUMBER_CHROMOSOMES+(i+1)*NUMBER_CHROMOSOMES+j] = current_population[(selector==0?parrent2:parrent1)*NUMBER_CHROMOSOMES+j];
    }
  }
}

int select_one_organism(double* fitness_values, double total_fitness, gsl_rng* r) {
  double running_total = 0;
  double selected_point = gsl_rng_uniform(r)*total_fitness;
  for(int i=0;i<POPULATION_SIZE;i++) {
    running_total+=fitness_values[i];
    if(running_total>selected_point) return i;
  }
  printf("Selection missed the wheel of fortune!, %f, %f\n", total_fitness, selected_point);
  return 0;
}

void apply_mutations(int* population, gsl_rng* r) {
  for(int i=0;i<POPULATION_SIZE;i++) {
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      if(gsl_rng_uniform(r) < MUTATION_RATE) {
        population[i*NUMBER_CHROMOSOMES+j] = generate_random_chromosome(r);
      }
    }
  }
}

void print_orgamism(int* population, int individ) {
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    printf("%d ", population[individ*NUMBER_CHROMOSOMES+i]);
  }
  printf(", %d\n", individ);
}

gsl_rng* create_rng(int seed, bool seed_time) {
  const gsl_rng_type * T;
  gsl_rng * r;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  struct timeval time;
  gettimeofday( &time, NULL );
  if(seed_time) {
    gsl_rng_set(r, seed*(int)time.tv_usec);
  } else {
    gsl_rng_set(r, seed);
  }
  return r;
}

int run() {
  gsl_rng** task_rngs = malloc(NUMBER_TASKS*sizeof(gsl_rng*));
  for(int i=0;i<NUMBER_TASKS;i++) {
    task_rngs[i] = create_rng(i+1, true);
  }
  int* current_population = initialize_population();
  int* next_population = malloc(POPULATION_SIZE*NUMBER_CHROMOSOMES*sizeof(int));
  double* fitness_values = malloc(POPULATION_SIZE*sizeof(double));
  double best_fitness = 0;
  int generation = 0;
  int task_ids[8] = {0,1,2,3,4,5,6,7};
  while(best_fitness<0.999999 && generation < MAXIMUM_GENERATIONS) {
    best_fitness = evaluate_organisms_gpu(current_population, fitness_values);
    for(int i=0;i<NUMBER_TASKS;i++) {
      produce_next_generation(current_population, next_population, fitness_values, task_ids[i], task_rngs[i]);
    }
    apply_mutations(next_population, task_rngs[0]);
    int* tmp_population = next_population;
    next_population = current_population;
    current_population = tmp_population;
    generation++;
    printf("generation, %d, best_fitness, %f\n", generation, best_fitness);
  }
  printf("Final generation: %d\n", generation);
  return generation;
}

void print_timestamp() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long long millisecondsSinceEpoch =
    (unsigned long long)(tv.tv_sec) * 1000 +
    (unsigned long long)(tv.tv_usec) / 1000;
  printf("%llu\n", millisecondsSinceEpoch);
}

int main() {
  print_timestamp();
  for(int i=0;i<RUNS;i++) {
    run();
    print_timestamp();
  }
  return 0;
}
