#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <gsl/gsl_rng.h>
#include "ga.h"
#include "omp.h"


// OpenCL imports.
#include <CL/cl.h>
#define MEM_SIZE (128)
#define MAX_SOURCE_SIZE (0x100000)

#define NUMBER_TASKS          8
#define POPULATION_SIZE       1600
#define TASK_POPULATION_SIZE  POPULATION_SIZE/NUMBER_TASKS
#define NUMBER_CHROMOSOMES    150
#define ALLELES               4
#define MUTATION_RATE         0.0035
#define MAXIMUM_GENERATIONS   10000
#define RUNS                  1

int* initialize_population() {
  const gsl_rng_type * T;
  gsl_rng * r;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  int* population = malloc(POPULATION_SIZE*NUMBER_CHROMOSOMES*sizeof(int));
  for(int i=0;i<POPULATION_SIZE;i++) {
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      population[i*NUMBER_CHROMOSOMES+j] = generate_random_chromosome(r);
    }
  }
  return population;
}

Organism* create_organism(gsl_rng* r) {
  Organism* organism = malloc(sizeof(Organism));
  organism->geneotype = malloc(NUMBER_CHROMOSOMES*sizeof(int));
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    organism->geneotype[i] = generate_random_chromosome(r);
  }
  organism->fitness = 0;
  return organism;
}

int generate_random_chromosome(gsl_rng* r) {
  return gsl_rng_uniform_int(r, ALLELES);
}

#pragma omp task in(population[0:factor*(task_size-1)]) out(fitness_values[0:task_size-1])
void evaluate_organisms(int* population, double* fitness_values, int factor, int task_size) {
  for(int i=0;i<POPULATION_SIZE;i++) {
    for(int j=0;j<1000;j++) {
      fitness_values[i] = evaluate_organism(population, i);
    }
  }
}

#pragma omp target device(opencl) ndrange(1, 1600, 1) implements(evaluate_organisms) file(ga.cl) copy_deps
#pragma omp task in(population[0:factor*(task_size-1)]) out(fitness_values[0:task_size-1])
void evaluate_organisms_gpu(int* population, double* fitness_values, int factor, int task_size);

double evaluate_organism(int* population, int individ) {
  double fitness;
  for(int j=0;j<1;j++) {
    int tally = 1;
    for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
      if(population[individ*NUMBER_CHROMOSOMES+i]!=1) tally++;
    }
    fitness = 1./(double)tally;
  }
  return fitness;
}

void produce_next_generation(int* current_population, int* next_population, double* fitness_values, int task_id, gsl_rng* r) {
  apply_crossover(current_population, next_population, fitness_values, task_id, r);
}

void apply_crossover(int* current_population, int* next_population, double* fitness_values, int task_id, gsl_rng* r) {
  double total_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    total_fitness+=fitness_values[i];
  }
  for(int i=0;i<POPULATION_SIZE/NUMBER_TASKS;i+=2) {
    int parrent1 = select_one_organism(fitness_values, total_fitness, r);
    int parrent2 = select_one_organism(fitness_values, total_fitness, r);
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      int selector = gsl_rng_uniform_int(r, 2);
      next_population[task_id*TASK_POPULATION_SIZE*NUMBER_CHROMOSOMES+i*NUMBER_CHROMOSOMES+j] = current_population[(selector==0?parrent1:parrent2)*NUMBER_CHROMOSOMES+j];
      next_population[task_id*TASK_POPULATION_SIZE*NUMBER_CHROMOSOMES+(i+1)*NUMBER_CHROMOSOMES+j] = current_population[(selector==0?parrent2:parrent1)*NUMBER_CHROMOSOMES+j];
    }
  }
}

int select_one_organism(double* fitness_values, double total_fitness, gsl_rng* r) {
  double running_total = 0;
  double selected_point = gsl_rng_uniform(r)*total_fitness;
  for(int i=0;i<POPULATION_SIZE;i++) {
    running_total+=fitness_values[i];
    if(running_total>selected_point) return i;
  }
  return POPULATION_SIZE-1;
}

void apply_mutations(int* population, gsl_rng* r) {
  for(int i=0;i<POPULATION_SIZE;i++) {
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      if(gsl_rng_uniform(r) < MUTATION_RATE) {
        population[i*NUMBER_CHROMOSOMES+j] = generate_random_chromosome(r);
      }
    }
  }
}

void print_orgamism(int* population, int individ) {
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    printf("%d ", population[individ*NUMBER_CHROMOSOMES+i]);
  }
  printf(", %d\n", individ);
}

gsl_rng* create_rng(int seed, bool seed_time) {
  const gsl_rng_type * T;
  gsl_rng * r;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  struct timeval time;
  gettimeofday( &time, NULL );
  if(seed_time) {
    gsl_rng_set(r, seed*(int)time.tv_usec);
  } else {
    gsl_rng_set(r, seed);
  }
  return r;
}

int run() {
  gsl_rng** task_rngs = malloc(NUMBER_TASKS*sizeof(gsl_rng*));
  for(int i=0;i<NUMBER_TASKS;i++) {
    task_rngs[i] = create_rng(i+1, true);
  }
  int* current_population = initialize_population();
  int* next_population = malloc(POPULATION_SIZE*NUMBER_CHROMOSOMES*sizeof(int));
  double* fitness_values = malloc(POPULATION_SIZE*sizeof(double));
  double best_fitness = 0;
  int generation = 0;
  int task_ids[8] = {0,1,2,3,4,5,6,7};
  while(best_fitness<0.999999 && generation < MAXIMUM_GENERATIONS) {
    for(int i=0;i<1;i++) {
      evaluate_organisms(current_population, fitness_values, NUMBER_CHROMOSOMES, POPULATION_SIZE);
    }
#pragma omp taskwait
    best_fitness = 0;
    for(int i=0;i<POPULATION_SIZE;i++) {
      if(best_fitness<fitness_values[i]) best_fitness = fitness_values[i];
    }
    for(int i=0;i<NUMBER_TASKS;i++) {
      produce_next_generation(current_population, next_population, fitness_values, task_ids[i], task_rngs[i]);
    }
    apply_mutations(next_population, task_rngs[0]);
    int* tmp_population = next_population;
    next_population = current_population;
    current_population = tmp_population;
    generation++;
  }
  printf("Final generation: %d\n", generation);
  return generation;
}

void print_timestamp() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long long millisecondsSinceEpoch =
    (unsigned long long)(tv.tv_sec) * 1000 +
    (unsigned long long)(tv.tv_usec) / 1000;
  printf("%llu\n", millisecondsSinceEpoch);
}

int main() {
  print_timestamp();
  for(int i=0;i<RUNS;i++) {
    run();
    print_timestamp();
  }
  return 0;
}
