#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <sys/time.h>
#include <gsl/gsl_rng.h>
#include "simple_ga.h"

#define POPULATION_SIZE     1600
#define NUMBER_CHROMOSOMES  150
#define ALLELES             4
#define MUTATION_RATE       0.0035
#define MAXIMUM_GENERATIONS 10000
#define RUNS                15

Organism*** initialize(gsl_rng* r) {
  Organism*** population_buffers = malloc(2*sizeof(Organism**));
  // First population buffer filled with random organisms
  population_buffers[0] = malloc(POPULATION_SIZE*sizeof(Organism*));
  for(int j=0;j<POPULATION_SIZE; j++) {
    population_buffers[0][j] = create_organism(r);
  }
  // Secod population buffer filled with empty organisms
  population_buffers[1] = malloc(POPULATION_SIZE*sizeof(Organism*));
  for(int i=0;i<POPULATION_SIZE;i++) {
    population_buffers[1][i] = malloc(sizeof(Organism));
    population_buffers[1][i]->geneotype = malloc(NUMBER_CHROMOSOMES*sizeof(int));
  }
  return population_buffers;
}

Organism* create_organism(gsl_rng* r) {
  Organism* organism = malloc(sizeof(Organism));
  organism->geneotype = malloc(NUMBER_CHROMOSOMES*sizeof(int));
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    organism->geneotype[i] = generate_random_chromosome(r);
  }
  organism->fitness = 0;
  return organism;
}

int generate_random_chromosome(gsl_rng* r) {
  return gsl_rng_uniform_int(r, ALLELES);
}

double evaluate_organisms(Organism** population) {
  double best_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    evaluate_organism(population[i]);
    if(best_fitness<population[i]->fitness) best_fitness = population[i]->fitness;
  }
  return best_fitness;
}

void evaluate_organism(Organism* organism) {
  int tally = 1;
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    if(organism->geneotype[i]!=1) tally++;
  }
  organism->fitness = 1./(double)tally;
}

Organism** produce_next_generation(Organism** current_population, Organism** next_population, gsl_rng* r) {
  apply_crossover(current_population, next_population, r);
  apply_mutations(next_population, r);
}

void apply_crossover(Organism** current_population, Organism** next_population, gsl_rng* r) {
  double total_fitness = 0;
  for(int i=0;i<POPULATION_SIZE;i++) {
    total_fitness+=current_population[i]->fitness;
  }
  for(int i=0;i<POPULATION_SIZE;i+=2) {
    int parrent1 = select_one_organism(current_population, total_fitness, r);
    int parrent2 = select_one_organism(current_population, total_fitness, r);
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      int selector = gsl_rng_uniform_int(r, 2);
      next_population[i]->geneotype[j] = current_population[selector==0?parrent1:parrent2]->geneotype[j];
      next_population[i+1]->geneotype[j] = current_population[selector==0?parrent2:parrent1]->geneotype[j];
    }
  }
}

int select_one_organism(Organism** population, double total_fitness,  gsl_rng* r) {
  double running_total = 0;
  double selected_point = gsl_rng_uniform(r)*total_fitness;
  for(int i=0;i<POPULATION_SIZE;i++) {
    running_total+=population[i]->fitness;
    if(running_total>selected_point) return i;
  }
  return 0;
}

void apply_mutations(Organism** population, gsl_rng* r) {
  for(int i=0;i<POPULATION_SIZE;i++) {
    for(int j=0;j<NUMBER_CHROMOSOMES;j++) {
      if(gsl_rng_uniform(r) < MUTATION_RATE) {
        population[i]->geneotype[j] = generate_random_chromosome(r);
      }
    }
  }
}

void print_orgamism(Organism* organism) {
  for(int i=0;i<NUMBER_CHROMOSOMES;i++) {
    printf("%i ", organism->geneotype[i]);
  }
  printf("[%f]\n", organism->fitness);
}

void run(gsl_rng* r) {
  Organism*** population_buffers = initialize(r);
  double best_fitness = 0;
  int generation = 0;
  Organism** current_population;
  Organism** next_population;
  while(generation<MAXIMUM_GENERATIONS && best_fitness<0.999999) {
    current_population = population_buffers[generation%2];
    next_population = population_buffers[(generation+1)%2];

    best_fitness = evaluate_organisms(current_population);
    produce_next_generation(current_population, next_population, r);
    generation++;
  }
  printf("Final generation: %d\n", generation);
}

void print_timestamp() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long long millisecondsSinceEpoch =
    (unsigned long long)(tv.tv_sec) * 1000 +
    (unsigned long long)(tv.tv_usec) / 1000;
  printf("%llu\n", millisecondsSinceEpoch);
}

gsl_rng* create_rng(int seed, bool seed_time) {
  const gsl_rng_type * T;
  gsl_rng * r;
  int i, n = 10;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  struct timeval time;
  gettimeofday( &time, NULL );
  if(seed_time) {
    gsl_rng_set(r, seed*(int)time.tv_usec);
  } else {
    gsl_rng_set(r, seed);
  }
  return r;
}

int main() {
  gsl_rng* r = create_rng(0,false);
  print_timestamp();
  for(int i=0;i<RUNS;i++) {
    run(r);
    print_timestamp();
  }
  return 0;
}
